// -*- Mode:C++ -*-
//Header:
//File: SimplePlayer.h
//Title: Simple Player for Soccer Server 7
//Author: NODA, Itsuki
//Date: 2001/03/10
//

//ModifyHistory:
// 2001/03/10: Start to create this file
//EndModifyHistory:

#ifndef _SimplePlayer_h_
#define _SimplePlayer_h_
////////////////////////////////////////////////////////////////////////

#include "itk/itk.h"
#include "sctypes.h"
#include "BasePlayer.h"
#include "estimatePos.h"
#include "FieldState.h"
#include "ThreadedPlayer.h"

namespace SClient {

   using namespace Itk ;
   using namespace Soccer ;

   //======================================================================
   // Simple Player

   class SimplePlayer : public ThreadedPlayer {
      public:
	 virtual const char * _classname() const { return "SimplePlayer";};

	 //==================================================
	 // body (SimplePlayer)

	 //--------------------------------------------------
	 // member
      public:

	 //------------------------------
	 // timestamps

	 CycleStep _stime ;	/* most recent sensor timestamp */
	 CycleStep _atime ;	/* most recent audio sensor timestamp */
	 //CycleStep _vtime ;   /* most recent visual sensor timestamp */
	 CycleStep _btime ;	/* most recent sense body timestamp */
	 CycleStep _fstime ;    /* most recent fullstate timestamp */

	 //------------------------------
	 // FieldState

	 FieldState _currentState ;	/* by sensor information */
	 FieldState _taughtState ;	/* by fullstate information */

	 //------------------------------
	 // major landmark

	 Landmark * _goal ;
	 Landmark & goal() { return *_goal ; } ;
	 Vec & goalpos() { return goal().pos ; } ;

	 Landmark * _owngoal ;
	 Landmark & owngoal() { return *_owngoal ; } ;
	 Vec & owngoalpos() { return owngoal().pos ; } ;

	 //------------------------------
	 // messages and playmode

	 MutexCondVar cvPlayMode ;

	 //--------------------------------------------------
	 // constructor
      public:
	 SimplePlayer() {
	    _goal = getLandmark(LMID_GR) ;
	    _owngoal = getLandmark(LMID_GL) ;
	 } ;
	 
	 
	 //--------------------------------------------------
	 // access

	 //------------------------------
	 // taught state by fullstate info
      public:
	 ITK_DEF_ACCESS(FieldState & taughtState(), {
	    return _taughtState ;
	 }) ;

	 //------------------------------
	 // field state
      public:
	 ITK_DEF_ACCESS(FieldState & currentState(), {
	    return _currentState ;
	 }) ;

	 ITK_DEF_ACCESS(CycleStep & stime(), { 	// sensor time
	    return _stime ;
	 });
	 ITK_DEF_ACCESS(CycleStep & atime(), { 	// hear time
	    return _atime ;
	 });
	 ITK_DEF_ACCESS(CycleStep & btime(), { 	// sense body time
	    return _btime ;
	 });
	 ITK_DEF_ACCESS(CycleStep & fstime(), { // full state time
	    return _fstime ;
	 });

	 ITK_DEF_ACCESS(CycleStep & vtime(), { 	// visual sensor time
	    return currentState().time();
	 });

	 ITK_DEF_ACCESS(PosDir & myposdir(), { 
	    return currentState().myposdir() ;
	 });
	 ITK_DEF_ACCESS(Vec & mypos(), { 
	    return currentState().mypos() ; 
	 });
	 ITK_DEF_ACCESS(Angle & myhdir(), { 
	    return currentState().myhdir() ; 
	 });
	 ITK_DEF_ACCESS(Angle & mybdir(), { 
	    return currentState().mybdir() ; 
	 });
	 ITK_DEF_ACCESS(Flt & posConf(), { 
	    return currentState().posConf() ;
	 });
	 ITK_DEF_ACCESS(Vec & myvel(), {
	    return currentState().myvel() ;
	 }) ;
	 ITK_DEF_ACCESS(Flt & velConf(), {
	    return currentState().velConf() ;
	 }) ;
	 ITK_DEF_ACCESS(Bool & outofFieldP(), {
	    return currentState().outofFieldP() ;
	 }) ;
	 ITK_DEF_ACCESS(FieldState::Object & ball(), { 
	    return currentState().ball() ;
	 }) ;
	 ITK_DEF_ACCESS(Vec & ballpos(), {
	    return currentState().ballpos() ;
	 }) ;
	 ITK_DEF_ACCESS(Vec & ballvel(), {
	    return currentState().ballvel() ;
	 }) ;

	 ITK_DEF_ACCESS(FieldState::Object & teammate(UInt i), {
	    return currentState().teammate(i) ; 
	 }) ;

	 ITK_DEF_ACCESS(FieldState::Object & opponent(UInt i), {
	    return currentState().opponent(i) ;
	 }) ;

	 ITK_DEF_ACCESS(FieldState::Object & player(UInt i), {
	    return currentState().player(i) ;
	 }) ;

	 //--------------------------------------------------
	 // analyze see/hear info
      public:

	 virtual Bool analyzeSensorSee() ;
	 virtual Bool analyzeSensorHear() ;
	 virtual Bool analyzeSensorSenseBody() ;
	 virtual Bool analyzeSensorFullState() ;

	 //--------------------------------------------------
	 // basic command
      public:

	 virtual Bool move(const Flt & x, const Flt & y) ;
	 Bool move(const Vec & pos) { return move(pos.x,pos.y) ; } ;

	 virtual Bool turn(const Flt & moment) ;
	 virtual Bool dash(const Flt & power) ;
	 virtual Bool kick(const Flt & power, const Flt & dir) ;
	 virtual Bool gcatch(const Flt & dir) ; //goalie catch

	 virtual Bool head(const Flt & dir) ;  // turn_neck

	 virtual Bool say(const SubString & msg) ;
	 Bool say(const char * msg) { 
	    return say(SubString(const_cast<char *>(msg))) ; 
	 } ;
	 
	 //--------------------------------------------------
	 // low level command

	 //------------------------------
	 // turn
      public:
	 virtual Bool turnTo(const Flt adir) ;
	 virtual Bool turnToRel(const Vec & rpos) ;
	 virtual Bool turnToAbs(const Vec & apos) ;

	 //------------------------------
	 // kick
      public:
	 virtual Bool kickTo(const Flt & power, const Flt & adir) ;
	 virtual Bool kickToRel(const Flt & power, const Vec & rpos) ;
	 virtual Bool kickToAbs(const Flt & power, const Vec & apos) ;

	 //------------------------------
	 // neck
      public:
	 virtual Bool headTo(const Angle & adir) ;
	 virtual Bool headToRel(const Vec & rpos) ;
	 virtual Bool headToAbs(const Vec & apos) ;
	 
	 //--------------------------------------------------
	 // timestamp check
      public:
	 static const Int SeenMargine = 5 ;

	 Bool isBallSeenRecently(Int margine = SeenMargine) const {
	    return ball().time().isNewer(stime(),margine) ;
	 } ;

	 //--------------------------------------------------
	 // kickable check
      public:
	 Bool kickable() const {
	    return ballpos().distance(mypos()) < SC_KICKABLE_AREA ;
	 } ;

	 //--------------------------------------------------
	 // mode check
      public:
	 Bool canMove() const {
	    if(isNull(playmode())) return False ;
	    switch(playmode()->id) {
	       case SC_WID_BeforeKickOff:
	       case SC_WID_HalfTime:
	       case SC_WID_GoalL:
	       case SC_WID_GoalR:
	       case SC_WID_OffSideL:
	       case SC_WID_OffSideR:
		  return True ;
	       default:
		  return False ;
	    }
	 } ;
	 
	 Bool isPlayOn() const {
	    switch(playmode()->id) {
	       case SC_WID_PlayOn:
		  return True ;
	       default:
		  return False ;
	    }
	 } ;

	 Bool isRestart(Side s) const {
	    if(isLeft(s)) {
	       switch(playmode()->id) {
		  case SC_WID_KickOffL:
		  case SC_WID_FreeKickL:
		  case SC_WID_CornerKickL:
		  case SC_WID_GoalKickL:
		  case SC_WID_GoalieCatchL:
		  case SC_WID_OffSideR:
		  case SC_WID_DropBall:
		     return True ;
		  default:
		     return False ;
	       }
	    } else {
	       switch(playmode()->id) {
		  case SC_WID_KickOffR:
		  case SC_WID_FreeKickR:
		  case SC_WID_CornerKickR:
		  case SC_WID_GoalKickR:
		  case SC_WID_GoalieCatchR:
		  case SC_WID_OffSideL:
		  case SC_WID_DropBall:
		     return True ;
		  default:
		     return False ;
	       }
	    }
	 } ;
	 
	 Bool isMyRestart() const { return isRestart(side) ; } ;
	 
	 Bool isOpponentRestart() const {
	    return isRestart(side) && (playmode()->id != SC_WID_DropBall) ;
	 } ;

	 Bool canPlay() const { return isPlayOn() || isMyRestart() ; } ;

	 Bool canDash() const { 
	    switch(playmode()->id) {
	       case SC_WID_BeforeKickOff:
	       case SC_WID_HalfTime:
		  return False ;
	       default:
		  return True ;
	    }
	 } ;

	 //--------------------------------------------------
	 // describe Aux
      public:
	 virtual void describeAux(ostream & ostr, 
				  const Bool detailp = True) const {
	    return ;
	 } ;
   } ;
} ;
////////////////////////////////////////////////////////////////////////
#endif
	 


   
